function ProcessImage( obj, event, himage )
% an update-event handler for the camera.
% this function will be invoked for each frame the camera captures
%
% 1.during the first %BTFrames% frames, this function will learn the static
%   background of the image plane
% 2.then the function will learn the controller (happens only once).
% 3.the third stage is to cut out all the non-moving non-skin pixels
%   this is achived by getting the foreground as a logical matrix (moving
%   pixel = 1, static pixel = 0) and the skin as a logical matrix (skin
%   pixel = 1, non-skin pixel = 0).
%   and we'll take their intersection by performing a logical AND.
    global isBackgroundModel E S A B;         %background detection values
    global K angle;                           %K-curvature constants
    global joystickRegions initJoystickRegions buttons ellipses;    %button detection values
    global pressed;
    if(~isBackgroundModel)
        LearnBackground(event.Data);
        return;
    end
    if (~initJoystickRegions)
        disp('Learning Controller ...');
        [joystickRegions, buttons,ellipses,~] = detectJoystickShapes(rgb2gray(event.Data));
        disp('Done!');
        disp('Start Clicking!');
        pressed = zeros(1,length(buttons));
        return;
    end
    cla(himage);                                                    %clears all plots
    im = double(event.Data);                                        %the current input frame from the camera
    foreground = classifyImage(im, E, S, A, B);                     %calculates the foreground mask
    skinned = (computeSkinProbability(im)>0);                       %calculates the skin mask
    [movingSkin,s] = enhanceHand(foreground, skinned);
    if (s == 0)
        set(himage, 'CData', event.Data);
        return;
    end
    boundary = bwtraceboundary(movingSkin,GetFirstNonZeroInBW(movingSkin),'N');
    tips = fingerTipsFromBoundary(boundary,K,angle);
    set(himage, 'CData', MaskOfRgb(event.Data,movingSkin));
    hold on;

    %plot the buttons
    plotButtons(joystickRegions, ellipses,buttons);
    %plot hand contuor
    plot(gca,boundary(:,2)',boundary(:,1)','LineWidth',2,'Color','red');
    %plot fingertips
    plotFingertips(tips)
    sizes = size(tips);
    newPressed = zeros(size(pressed));
    for i= 1:sizes(1)
        idx = buttonFromPoint(tips(i,1),tips(i,2));
        if(idx ~= 0)
            btn = buttons(2,idx);
            if (btn ~=0)
                if (pressed(idx) == 0 && newPressed(idx) == 0)
                    DoAction(btn);
                end
                newPressed(idx) = 1;
            end
        end
    end
    pressed = newPressed;
end

%%%%%%%%%===================== inner methods =====================%%%%%%%%%
%%

function DoAction(buttonID)
    global notepad;
    keys = '';
    if(buttonID == 1)
        keys = 'Right ';
    elseif(buttonID == 2)
        keys = 'Left ';
    elseif(buttonID == 3)
        keys = 'Down ';
    elseif(buttonID == 4)
        keys = 'Up ';
    elseif(buttonID == 5)
        keys = 'Space ';
    elseif(buttonID == 6)
        keys = 'World! ';
    elseif(buttonID == 7)
        keys = 'O ';
    elseif(buttonID == 8)
        keys = 'L';
    elseif(buttonID == 9)
        keys = 'E';
    elseif(buttonID == 10)
        keys = 'H';
    end
    notepad.SendKeys(keys); %Sends keystrokes
end

%%
function [movingSkin,s] = enhanceHand(foreground, skinned)
    SE = strel('disk',1,4);
    foreground = imdilate(foreground,SE);                           %enhance foreground
    skinned = imdilate(imdilate(imdilate(skinned,SE),SE),SE);       %enhance skin 
    movingSkin = GetMainConnectedComponent(skinned & foreground);   %removes all image parts that doesnt belong to the main CC
    props = regionprops(movingSkin,'BoundingBox','FilledImage','Area');
    sizes = size(props);
    if (sizes(1) ~= 0)
        if (props.Area < 150)
            s = 0;
            return;    
        end
        s = 1;
        dim = props.BoundingBox;
        movingSkin(ceil(dim(2)):(ceil(dim(2))+floor(dim(4)) - 1),ceil(dim(1)):(ceil(dim(1))+floor(dim(3)) - 1)) = props.FilledImage;
    else
        s = 0;
        return;
    end
end

%%
function rgb = MaskOfRgb(im,BW)
%multiply each pixel by the value of the mask to get the masked image
    BW = uint8(BW);
    tmpR = im(:,:,1).*BW;
    tmpG = im(:,:,2).*BW;
    tmpB = im(:,:,3).*BW;
    rgb = cat(3,cat(3,tmpR,tmpG), tmpB);
end
%%
function plotButtons(JoystickRegions, ellipses,buttons)
%plots the button recognized in the controller
    for i = 1:length(JoystickRegions)
        hull=JoystickRegions(i).ConvexHull;
        color = 'green';
        width = 1;
        if (ismember(i,ellipses))
            color = 'red';
        end
        if (ismember(i,buttons))
            width = 3;
        end
        plot(hull(:,1)',hull(:,2)','LineWidth',width,'Color',color);
    end
end
%%
function plotFingertips(tips)
%plots a line between (r,c)=(0,0) and each finget tip
    sizes = size(tips);
    for i=1:sizes(1)
        plot(gca,[0;tips(i,2)],[0,tips(i,1)],'LineWidth',1,'Color','blue');
    end
end
%%
function LearnBackground(visibleImage)
% In this background training process, parameters associated with
% normalization are computed over N number of static background frames.
% The background is modeled statistically on a pixel by pixel basis, with
% each pixel being modeled by a 4-tuple <E, s, a, b>.
% E is the expected color value
% S is the standard deviation of color value
% A is the variation of the brightness distortion
% B is the variation of the chromaticity distortion
    global isBackgroundModel;
    global BackgroundFramesSequence;
    global frameCount BTFrames E S A B;
    if (frameCount == 0)
        disp('Learning Backgorund ...');
        BackgroundFramesSequence = visibleImage;
        frameCount = frameCount + 1;
    else        
        BackgroundFramesSequence = cat(4,BackgroundFramesSequence, visibleImage);
        frameCount = frameCount + 1;
        if (frameCount == BTFrames)
            disp('Done!');
            disp('Calculating Background Values ...');
            BackgroundFramesSequence = double(BackgroundFramesSequence);
            % Calculate the mean values for each pixel in the background model
            E = mean(BackgroundFramesSequence, 4);

            % Calculate the standard deviation for each pixel in the background model
            S = std(BackgroundFramesSequence, 1, 4);

            tempA = zeros(size(BackgroundFramesSequence(:,:,1,1)));
            tempB = zeros(size(BackgroundFramesSequence(:,:,1,1)));

            for i=1:size(BackgroundFramesSequence,4)
               alpha = brightnessDistortion(BackgroundFramesSequence(:,:,:,i), E, S);
               CD = colorDistortion(BackgroundFramesSequence(:,:,:,i), E, S, alpha);

               tempA = tempA + ((alpha - ones(size(alpha))) .^ 2);
               tempB = tempB + (CD .^ 2);
            end;

            A = (tempA / size(BackgroundFramesSequence,4)) .^ (1/2);
            B = (tempB / size(BackgroundFramesSequence,4)) .^ (1/2);

            isBackgroundModel = true;
            disp('Done!');
            disp('Tracking Starting ...');
        end
    end
end
%%
function p = GetFirstNonZeroInBW( BW )
%a dumb utility method
%we needed a 1-valued cell coords, any set of correct coords.. 
%so we implemented the dumbest search to get one
    p = size(BW);
    for i=1:p(1)
        for j=1:p(2)
            if (BW(i,j)==1)
                p=[i,j];
                break;
            end
        end
    end
end
%%
function newBW = GetMainConnectedComponent(BW)
%calculates the varius connected componnets in the BW
%and return the biggest one laced on same-size-as-BW zeros matrix
    CC = bwconncomp(BW);
    numPixels = cellfun(@numel,CC.PixelIdxList);
    [~,idx] = max(numPixels);
    clear numPixels;
    newBW = uint8(zeros(size(BW)));
    newBW(CC.PixelIdxList{idx}) = 1;
end


